# Collect facts related to system service manager and init.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import platform
import re
import typing as t

from ansible.module_utils.common.text.converters import to_native

from ansible.module_utils.facts.utils import get_file_content
from ansible.module_utils.facts.collector import BaseFactCollector

# The distutils module is not shipped with SUNWPython on Solaris.
# It's in the SUNWPython-devel package which also contains development files
# that don't belong on production boxes.  Since our Solaris code doesn't
# depend on LooseVersion, do not import it on Solaris.
if platform.system() != 'SunOS':
    from ansible.module_utils.compat.version import LooseVersion


class ServiceMgrFactCollector(BaseFactCollector):
    name = 'service_mgr'
    _fact_ids = set()  # type: t.Set[str]
    required_facts = set(['platform', 'distribution'])

    @staticmethod
    def is_systemd_managed(module):
        # tools must be installed
        if module.get_bin_path('systemctl'):

            # this should show if systemd is the boot init system, if checking init failed to mark as systemd
            # these mirror systemd's own sd_boot test http://www.freedesktop.org/software/systemd/man/sd_booted.html
            for canary in ["/run/systemd/system/", "/dev/.run/systemd/", "/dev/.systemd/"]:
                if os.path.exists(canary):
                    return True
        return False

    @staticmethod
    def is_systemd_managed_offline(module):
        # tools must be installed
        if module.get_bin_path('systemctl'):
            # check if /sbin/init is a symlink to systemd
            # on SUSE, /sbin/init may be missing if systemd-sysvinit package is not installed.
            if os.path.islink('/sbin/init') and os.path.basename(os.readlink('/sbin/init')) == 'systemd':
                return True
        return False

    def collect(self, module=None, collected_facts=None):
        facts_dict = {}

        if not module:
            return facts_dict

        collected_facts = collected_facts or {}
        service_mgr_name = None

        # TODO: detect more custom init setups like bootscripts, dmd, s6, Epoch, etc
        # also other OSs other than linux might need to check across several possible candidates

        # Mapping of proc_1 values to more useful names
        proc_1_map = {
            'procd': 'openwrt_init',
            'runit-init': 'runit',
            'svscan': 'svc',
            'openrc-init': 'openrc',
        }

        # try various forms of querying pid 1
        proc_1 = get_file_content('/proc/1/comm')
        if proc_1 is None:
            rc, proc_1, err = module.run_command("ps -p 1 -o comm|tail -n 1", use_unsafe_shell=True)

            # if command fails, or stdout is empty string or the output of the command starts with what looks like a PID,
            # then the 'ps' command probably didn't work the way we wanted, probably because it's busybox
            if rc != 0 or not proc_1.strip() or re.match(r' *[0-9]+ ', proc_1):
                proc_1 = None

        # The ps command above may return "COMMAND" if the user cannot read /proc, e.g. with grsecurity
        if proc_1 == "COMMAND\n":
            proc_1 = None

        if proc_1 is None and os.path.islink('/sbin/init'):
            proc_1 = os.readlink('/sbin/init')

        if proc_1 is not None:
            proc_1 = os.path.basename(proc_1)
            proc_1 = to_native(proc_1)
            proc_1 = proc_1.strip()

        if proc_1 is not None and (proc_1 == 'init' or proc_1.endswith('sh')):
            # many systems return init, so this cannot be trusted, if it ends in 'sh' it probably is a shell in a container
            proc_1 = None

        # if not init/None it should be an identifiable or custom init, so we are done!
        if proc_1 is not None:
            # Lookup proc_1 value in map and use proc_1 value itself if no match
            service_mgr_name = proc_1_map.get(proc_1, proc_1)

        # start with the easy ones
        elif collected_facts.get('ansible_distribution', None) == 'MacOSX':
            # FIXME: find way to query executable, version matching is not ideal
            if LooseVersion(platform.mac_ver()[0]) >= LooseVersion('10.4'):
                service_mgr_name = 'launchd'
            else:
                service_mgr_name = 'systemstarter'
        elif 'BSD' in collected_facts.get('ansible_system', '') or collected_facts.get('ansible_system') in ['Bitrig', 'DragonFly']:
            # FIXME: we might want to break out to individual BSDs or 'rc'
            service_mgr_name = 'bsdinit'
        elif collected_facts.get('ansible_system') == 'AIX':
            service_mgr_name = 'src'
        elif collected_facts.get('ansible_system') == 'SunOS':
            service_mgr_name = 'smf'
        elif collected_facts.get('ansible_distribution') == 'OpenWrt':
            service_mgr_name = 'openwrt_init'
        elif collected_facts.get('ansible_distribution') == 'SMGL':
            service_mgr_name = 'simpleinit_msb'
        elif collected_facts.get('ansible_system') == 'Linux':
            # FIXME: mv is_systemd_managed
            if self.is_systemd_managed(module=module):
                service_mgr_name = 'systemd'
            elif module.get_bin_path('initctl') and os.path.exists("/etc/init/"):
                service_mgr_name = 'upstart'
            elif os.path.exists('/sbin/openrc'):
                service_mgr_name = 'openrc'
            elif self.is_systemd_managed_offline(module=module):
                service_mgr_name = 'systemd'
            elif os.path.exists('/etc/init.d/'):
                service_mgr_name = 'sysvinit'
            elif os.path.exists('/etc/dinit.d/'):
                service_mgr_name = 'dinit'

        if not service_mgr_name:
            # if we cannot detect, fallback to generic 'service'
            service_mgr_name = 'service'

        facts_dict['service_mgr'] = service_mgr_name
        return facts_dict



from __future__ import annotations

import datetime
import functools
import json
import typing as t

from ansible.module_utils import _internal
from ansible.module_utils._internal import _messages
from ansible.module_utils._internal._datatag import (
    AnsibleSerializable,
    AnsibleSerializableWrapper,
    AnsibleTaggedObject,
    Tripwire,
    _AnsibleTaggedBytes,
    _AnsibleTaggedDate,
    _AnsibleTaggedDateTime,
    _AnsibleTaggedDict,
    _AnsibleTaggedFloat,
    _AnsibleTaggedInt,
    _AnsibleTaggedList,
    _AnsibleTaggedSet,
    _AnsibleTaggedStr,
    _AnsibleTaggedTime,
    _AnsibleTaggedTuple,
    AnsibleTagHelper,
    _tags,
)

# transformations to "final" JSON representations can only use:
# str, float, int, bool, None, dict, list
# NOT SUPPORTED: tuple, set -- the representation of these in JSON varies by profile (can raise an error, may be converted to list, etc.)
# This means that any special handling required on JSON types that are not wrapped/tagged must be done in a pre-pass before serialization.
# The final type map cannot contain any JSON types other than tuple or set.


_NoneType: t.Final[type] = type(None)

_json_subclassable_scalar_types: t.Final[tuple[type, ...]] = (str, float, int)
"""Scalar types understood by JSONEncoder which can also be subclassed."""

_json_scalar_types: t.Final[tuple[type, ...]] = (str, float, int, bool, _NoneType)
"""Scalar types understood by JSONEncoder."""

_json_container_types: t.Final[tuple[type, ...]] = (dict, list, tuple)
"""Container types understood by JSONEncoder."""

_json_types: t.Final[tuple[type, ...]] = _json_scalar_types + _json_container_types
"""Types understood by JSONEncoder."""

_intercept_containers = frozenset(
    {
        dict,
        list,
        tuple,
        _AnsibleTaggedDict,
        _AnsibleTaggedList,
        _AnsibleTaggedTuple,
    }
)
"""Container types to intercept in support of scalar interception."""

_common_module_types: frozenset[type[AnsibleSerializable]] = frozenset(
    {
        _AnsibleTaggedBytes,
        _AnsibleTaggedDate,
        _AnsibleTaggedDateTime,
        _AnsibleTaggedDict,
        _AnsibleTaggedFloat,
        _AnsibleTaggedInt,
        _AnsibleTaggedList,
        _AnsibleTaggedSet,
        _AnsibleTaggedStr,
        _AnsibleTaggedTime,
        _AnsibleTaggedTuple,
    }
)
"""
Types that must be supported for all Ansible module serialization profiles.

For module-to-controller, all types should support full fidelity serialization.
This allows infrastructure and library code to use these features even when a module does not.

For controller-to-module, type behavior is profile dependent.
"""

_common_module_response_types: frozenset[type[AnsibleSerializable]] = frozenset(
    {
        _messages.PluginInfo,
        _messages.PluginType,
        _messages.Event,
        _messages.EventChain,
        _messages.ErrorSummary,
        _messages.WarningSummary,
        _messages.DeprecationSummary,
        _tags.Deprecated,
    }
)
"""Types that must be supported for all Ansible module-to-controller serialization profiles."""

_T_encoder = t.TypeVar('_T_encoder', bound="AnsibleProfileJSONEncoder")
_T_decoder = t.TypeVar('_T_decoder', bound="AnsibleProfileJSONDecoder")


class _JSONSerializationProfile(t.Generic[_T_encoder, _T_decoder]):
    serialize_map: t.ClassVar[dict[type, t.Callable]]
    """
    Each concrete non-JSON type must be included in this mapping to support serialization.
    Including a JSON type in the mapping allows for overriding or disabling of serialization of that type.
    """

    deserialize_map: t.ClassVar[dict[str, t.Callable]]
    """A mapping of type keys to type dispatchers for deserialization."""

    allowed_ansible_serializable_types: t.ClassVar[frozenset[type[AnsibleSerializable]]] = frozenset()
    """Each concrete AnsibleSerialiable derived type must be included in this set to support serialization."""

    _common_discard_tags: t.ClassVar[dict[type, t.Callable]]
    """
    Serialize map for tagged types to have their tags discarded.
    This is generated by __init_subclass__ and should not be manually updated.
    """

    _allowed_type_keys: t.ClassVar[frozenset[str]]
    """
    The set of type keys allowed during deserialization.
    This is generated by __init_subclass__ and should not be manually updated.
    """

    _unwrapped_json_types: t.ClassVar[frozenset[type]]
    """
    The set of types that do not need to be wrapped during serialization.
    This is generated by __init_subclass__ and should not be manually updated.
    """

    profile_name: t.ClassVar[str]
    """
    The user-facing name of the profile, derived from the module name in which the profile resides.
    Used to load the profile dynamically at runtime.
    This is generated by __init_subclass__ and should not be manually updated.
    """

    encode_strings_as_utf8: t.ClassVar[bool] = False
    r"""
    When enabled, JSON encoding will result in UTF8 strings being emitted.
    Otherwise, non-ASCII strings will be escaped with `\uXXXX` escape sequences.`
    """

    @classmethod
    def pre_serialize(cls, encoder: _T_encoder, o: t.Any) -> t.Any:
        return o

    @classmethod
    def post_deserialize(cls, decoder: _T_decoder, o: t.Any) -> t.Any:
        return o

    @classmethod
    def cannot_serialize_error(cls, target: t.Any, /) -> t.NoReturn:
        raise TypeError(f'Object of type {type(target).__name__!r} is not JSON serializable by the {cls.profile_name!r} profile.')

    @classmethod
    def cannot_deserialize_error(cls, target_type_name: str, /) -> t.NoReturn:
        raise TypeError(f'Object of type {target_type_name!r} is not JSON deserializable by the {cls.profile_name!r} profile.')

    @classmethod
    def unsupported_target_type_error(cls, target_type_name: str, _value: dict) -> t.NoReturn:
        cls.cannot_deserialize_error(target_type_name)

    @classmethod
    def discard_tags(cls, value: AnsibleTaggedObject) -> object:
        return value._native_copy()

    @classmethod
    def deserialize_serializable(cls, value: dict[str, t.Any]) -> object:
        type_key = value[AnsibleSerializable._TYPE_KEY]

        if type_key not in cls._allowed_type_keys:
            cls.cannot_deserialize_error(type_key)

        return AnsibleSerializable._deserialize(value)

    @classmethod
    def serialize_as_list(cls, value: t.Iterable) -> list:
        # DTFIX-FUTURE: once we have separate control/data channels for module-to-controller (and back), warn about this conversion
        return AnsibleTagHelper.tag_copy(value, (item for item in value), value_type=list)

    @classmethod
    def serialize_as_isoformat(cls, value: datetime.date | datetime.time | datetime.datetime) -> str:
        return value.isoformat()

    @classmethod
    def serialize_serializable_object(cls, value: AnsibleSerializable) -> t.Any:
        return value._serialize()

    @classmethod
    def post_init(cls) -> None:
        pass

    @classmethod
    def maybe_wrap(cls, o: t.Any) -> t.Any:
        if type(o) in cls._unwrapped_json_types:
            return o

        return _WrappedValue(o)

    @classmethod
    def handle_key(cls, k: t.Any) -> t.Any:
        """Validation/conversion hook before a dict key is serialized. The default implementation only accepts str-typed keys."""
        # NOTE: Since JSON requires string keys, there is no support for preserving tags on dictionary keys during serialization.

        if not isinstance(k, str):  # DTFIX-FUTURE: optimize this to use all known str-derived types in type map / allowed types
            raise TypeError(f'Key of type {type(k).__name__!r} is not JSON serializable by the {cls.profile_name!r} profile.')

        return k

    @classmethod
    def _handle_key_str_fallback(cls, k: t.Any) -> t.Any:
        """Legacy implementations should use this key handler for backward compatibility with stdlib JSON key conversion quirks."""
        # DTFIX-FUTURE: optimized exact-type table lookup first

        if isinstance(k, str):
            return k

        if k is None or isinstance(k, (int, float)):
            return json.dumps(k)

        raise TypeError(f'Key of type {type(k).__name__!r} is not JSON serializable by the {cls.profile_name!r} profile.')

    @classmethod
    def default(cls, o: t.Any) -> t.Any:
        # Preserve the built-in JSON encoder support for subclasses of scalar types.

        if isinstance(o, _json_subclassable_scalar_types):
            return o

        # Preserve the built-in JSON encoder support for subclasses of dict and list.
        # Additionally, add universal support for mappings and sequences/sets by converting them to dict and list, respectively.

        if _internal.is_intermediate_mapping(o):
            return {cls.handle_key(k): cls.maybe_wrap(v) for k, v in o.items()}

        if _internal.is_intermediate_iterable(o):
            return [cls.maybe_wrap(v) for v in o]

        return cls.last_chance(o)

    @classmethod
    def last_chance(cls, o: t.Any) -> t.Any:
        if isinstance(o, Tripwire):
            o.trip()

        cls.cannot_serialize_error(o)

    def __init_subclass__(cls, **kwargs) -> None:
        cls.deserialize_map = {}
        cls._common_discard_tags = {obj: cls.discard_tags for obj in _common_module_types if issubclass(obj, AnsibleTaggedObject)}

        cls.post_init()

        cls.profile_name = cls.__module__.rsplit('.', maxsplit=1)[-1].lstrip('_')

        wrapper_types = set(obj for obj in cls.serialize_map.values() if isinstance(obj, type) and issubclass(obj, AnsibleSerializableWrapper))

        cls.allowed_ansible_serializable_types |= wrapper_types

        # no current need to preserve tags on controller-only types or custom behavior for anything in `allowed_serializable_types`
        cls.serialize_map.update({obj: cls.serialize_serializable_object for obj in cls.allowed_ansible_serializable_types})
        cls.serialize_map.update({obj: func for obj, func in _internal.get_controller_serialize_map().items() if obj not in cls.serialize_map})

        cls.deserialize_map[AnsibleSerializable._TYPE_KEY] = cls.deserialize_serializable  # always recognize tagged types

        cls._allowed_type_keys = frozenset(obj._type_key for obj in cls.allowed_ansible_serializable_types)

        cls._unwrapped_json_types = frozenset(
            {obj for obj in cls.serialize_map if not issubclass(obj, _json_types)}  # custom types that do not extend JSON-native types
            | {obj for obj in _json_scalar_types if obj not in cls.serialize_map}  # JSON-native scalars lacking custom handling
        )


class _WrappedValue:
    __slots__ = ('wrapped',)

    def __init__(self, wrapped: t.Any) -> None:
        self.wrapped = wrapped


class AnsibleProfileJSONEncoder(json.JSONEncoder):
    """Profile based JSON encoder capable of handling Ansible internal types."""

    _wrap_container_types = (list, set, tuple, dict)
    _profile: type[_JSONSerializationProfile]

    profile_name: str

    def __init__(self, **kwargs):
        self._wrap_types = self._wrap_container_types + (AnsibleSerializable,)

        if self._profile.encode_strings_as_utf8:
            kwargs.update(ensure_ascii=False)

        super().__init__(**kwargs)

    def __init_subclass__(cls, **kwargs) -> None:
        cls.profile_name = cls._profile.profile_name

    def encode(self, o):
        o = self._profile.maybe_wrap(self._profile.pre_serialize(self, o))

        return super().encode(o)

    def default(self, o: t.Any) -> t.Any:
        o_type = type(o)

        if o_type is _WrappedValue:  # pylint: disable=unidiomatic-typecheck
            o = o.wrapped
            o_type = type(o)

        if mapped_callable := self._profile.serialize_map.get(o_type):
            return self._profile.maybe_wrap(mapped_callable(o))

        # This is our last chance to intercept the values in containers, so they must be wrapped here.
        # Only containers natively understood by the built-in JSONEncoder are recognized, since any other container types must be present in serialize_map.

        if o_type is dict:  # pylint: disable=unidiomatic-typecheck
            return {self._profile.handle_key(k): self._profile.maybe_wrap(v) for k, v in o.items()}

        if o_type is list or o_type is tuple:  # pylint: disable=unidiomatic-typecheck
            return [self._profile.maybe_wrap(v) for v in o]  # JSONEncoder converts tuple to a list, so just make it a list now

        # Any value here is a type not explicitly handled by this encoder.
        # The profile default handler is responsible for generating an error or converting the value to a supported type.

        return self._profile.default(o)


class AnsibleProfileJSONDecoder(json.JSONDecoder):
    """Profile based JSON decoder capable of handling Ansible internal types."""

    _profile: type[_JSONSerializationProfile]

    profile_name: str

    def __init__(self, **kwargs):
        kwargs.update(object_hook=self.object_hook)

        super().__init__(**kwargs)

    def __init_subclass__(cls, **kwargs) -> None:
        cls.profile_name = cls._profile.profile_name

    def raw_decode(self, s: str, idx: int = 0) -> tuple[t.Any, int]:
        obj, end = super().raw_decode(s, idx)

        if _string_encoding_check_enabled():
            try:
                _recursively_check_string_encoding(obj)
            except UnicodeEncodeError as ex:
                raise _create_encoding_check_error() from ex

        obj = self._profile.post_deserialize(self, obj)

        return obj, end

    def object_hook(self, pairs: dict[str, object]) -> object:
        if _string_encoding_check_enabled():
            try:
                for key, value in pairs.items():
                    key.encode()
                    _recursively_check_string_encoding(value)
            except UnicodeEncodeError as ex:
                raise _create_encoding_check_error() from ex

        for mapped_key, mapped_callable in self._profile.deserialize_map.items():
            if mapped_key in pairs:
                return mapped_callable(pairs)

        return pairs


_check_encoding_setting = 'MODULE_STRICT_UTF8_RESPONSE'
r"""
The setting to control whether strings are checked to verify they can be encoded as valid UTF8.
This is currently only used during deserialization, to prevent string values from entering the controller which will later fail to be encoded as bytes.

The encoding failure can occur when the string represents one of two kinds of values:
1) It was created through decoding bytes with the `surrogateescape` error handler, and that handler is not being used when encoding.
2) It represents an invalid UTF8 value, such as `"\ud8f3"` in a JSON payload. This cannot be encoded, even using the `surrogateescape` error handler.

Although this becomes an error during deserialization, there are other opportunities for these values to become strings within Ansible.
Future code changes should further restrict bytes to string conversions to eliminate use of `surrogateescape` where appropriate.
Additional warnings at other boundaries may be needed to give users an opportunity to resolve the issues before they become errors.
"""
# DTFIX-FUTURE: add strict UTF8 string encoding checking to serialization profiles (to match the checks performed during deserialization)
# DTFIX3: the surrogateescape note above isn't quite right, for encoding use surrogatepass, which does work
# DTFIX-FUTURE: this config setting should probably be deprecated


def _create_encoding_check_error() -> Exception:
    """
    Return an AnsibleError for use when a UTF8 string encoding check has failed.
    These checks are only performed in the controller context, but since this is module_utils code, dynamic loading of the `errors` module is required.
    """
    errors = _internal.import_controller_module('ansible.errors')  # bypass AnsiballZ import scanning

    return errors.AnsibleRuntimeError(
        message='Refusing to deserialize an invalid UTF8 string value.',
        help_text=f'This check can be disabled with the `{_check_encoding_setting}` setting.',
    )


@functools.lru_cache
def _string_encoding_check_enabled() -> bool:
    """Return True if JSON deserialization should verify strings can be encoded as valid UTF8."""
    if constants := _internal.import_controller_module('ansible.constants'):  # bypass AnsiballZ import scanning
        return constants.config.get_config_value(_check_encoding_setting)  # covers all profile-based deserializers, not just modules

    return False


def _recursively_check_string_encoding(value: t.Any) -> None:
    """Recursively check the given object to ensure all strings can be encoded as valid UTF8."""
    value_type = type(value)

    if value_type is str:
        value.encode()
    elif value_type is list:  # dict is handled by the JSON deserializer
        for item in value:
            _recursively_check_string_encoding(item)



#!actually_python

from __future__ import annotations

import sys
import json


def main():
    print(json.dumps(dict(
        running_python_interpreter=sys.executable,
    )))


if __name__ == '__main__':
    main()



from __future__ import annotations

from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    """Specialized config lookup that applies data transformations on values that config cannot."""

    def run(self, terms, variables=None, **kwargs):
        if not terms or not (config_name := terms[0]):
            raise ValueError("config name is required")

        match config_name:
            case 'DISPLAY_TRACEBACK':
                # since config can't expand this yet, we need the post-processed version
                from ansible.module_utils._internal._traceback import traceback_for

                return traceback_for()
            # DTFIX-FUTURE: plumb through normal config fallback
            case _:
                raise ValueError(f"Unknown config name {config_name!r}.")



"""Test the `sys.intern` patch."""

from __future__ import annotations

import sys

from ansible.module_utils._internal._datatag._tags import Deprecated


def test_sys_intern() -> None:
    """Verify that `sys.intern` works with a tagged value."""
    assert sys.intern(Deprecated(msg='').tag('x')) == 'x'


def test_reference_equality() -> None:
    """Verify that `sys.intern` returns identical string references for the same string value."""
    class CustomStr(str):
        pass

    custom_str = CustomStr('hello')
    plain_str = "hello"

    assert sys.intern(custom_str) is sys.intern(plain_str)



"""Patches for the builtin `sys` module."""

from __future__ import annotations

import contextlib
import sys
import typing as t

from . import CallablePatch


class _CustomStr(str):
    """Wrapper around `str` to test if subclasses are accepted."""


class SysInternPatch(CallablePatch):
    """Patch `sys.intern` so that subclasses of `str` are accepted."""

    target_container: t.ClassVar = sys
    target_attribute = 'intern'

    @classmethod
    def is_patch_needed(cls) -> bool:
        with contextlib.suppress(TypeError):
            sys.intern(_CustomStr("x"))
            return False

        return True

    def __call__(self, value: str):
        if type(value) is not str and isinstance(value, str):  # pylint: disable=unidiomatic-typecheck
            value = str(value)

        return type(self).unpatched_implementation(value)



#!/usr/bin/env python

from __future__ import annotations

import argparse
import json
import os
import pathlib
import re
import sys


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('stdout', type=pathlib.Path)
    parser.add_argument('stderr', type=pathlib.Path)
    parser.add_argument('--match-stdout', type=re.compile, action='extend', default=[], help='matching regex patterns to save to stdout')
    parser.add_argument('--match-stderr', type=re.compile, action='extend', default=[], help='matching regex patterns to save to stderr')

    args = parser.parse_args()

    stdout_path: pathlib.Path = args.stdout
    stderr_path: pathlib.Path = args.stderr
    stdout_patterns: list[re.Pattern] = args.match_stdout
    stderr_patterns: list[re.Pattern] = args.match_stderr

    stdout_lines: list[str] = []
    stderr_lines: list[str] = []

    ansible_test_temp_root = os.environ['JUNIT_TASK_RELATIVE_PATH']

    for line in sys.stdin:
        line = re.sub(r'\033\[[\d+;]+m', '', line)  # remove ANSI color sequences
        line = re.sub(re.escape(ansible_test_temp_root), 'ANSIBLE_TEST_TEMP_ROOT', line)

        if (match := re.search(r'^(?P<prefix>.*? => )(?P<json>\{.+})$', line)) and (data := json.loads(match.group('json'))):
            stdout_lines.append(f'{match.group("prefix")}{json.dumps(data, indent=4)}\n')
        elif re.search(r'^(?:PLAYBOOK:|PLAY|TASK) ', line):
            stdout_lines.append(line)
        elif re.search(r'^\[(?:DEPRECATION WARNING|WARNING|ERROR)]: .*$', line):
            stderr_lines.append(line)
        elif any(pattern.match(line) for pattern in stdout_patterns):
            stdout_lines.append(line)
        elif any(pattern.match(line) for pattern in stderr_patterns):
            stderr_lines.append(line)
        else:
            continue

    stdout_path.write_text(''.join(stdout_lines))
    stderr_path.write_text(''.join(stderr_lines))


if __name__ == '__main__':
    main()



"""Layout provider for Ansible collections."""

from __future__ import annotations

import os

from . import (
    ContentLayout,
    LayoutProvider,
    CollectionDetail,
    LayoutMessages,
)

from ...util import (
    is_valid_identifier,
)


class CollectionLayout(LayoutProvider):
    """Layout provider for Ansible collections."""

    @staticmethod
    def is_content_root(path: str) -> bool:
        """Return True if the given path is a content root for this provider."""
        if os.path.basename(os.path.dirname(os.path.dirname(path))) == 'ansible_collections':
            return True

        return False

    def create(self, root: str, paths: list[str]) -> ContentLayout:
        """Create a Layout using the given root and paths."""
        plugin_paths = dict((p, os.path.join('plugins', p)) for p in self.PLUGIN_TYPES)

        collection_root = os.path.dirname(os.path.dirname(root))
        collection_dir = os.path.relpath(root, collection_root)

        collection_namespace: str
        collection_name: str

        collection_namespace, collection_name = collection_dir.split(os.path.sep)

        collection_root = os.path.dirname(collection_root)

        sanity_messages = LayoutMessages()
        integration_messages = LayoutMessages()
        unit_messages = LayoutMessages()

        # these apply to all test commands
        self.__check_test_path(paths, sanity_messages)
        self.__check_test_path(paths, integration_messages)
        self.__check_test_path(paths, unit_messages)

        # these apply to specific test commands
        integration_targets_path = self.__check_integration_path(paths, integration_messages)
        self.__check_unit_path(paths, unit_messages)

        errors: list[str] = []

        if not is_valid_identifier(collection_namespace):
            errors.append(f'The namespace "{collection_namespace}" is an invalid identifier or a reserved keyword.')

        if not is_valid_identifier(collection_name):
            errors.append(f'The name "{collection_name}" is an invalid identifier or a reserved keyword.')

        return ContentLayout(
            root,
            paths,
            plugin_paths=plugin_paths,
            collection=CollectionDetail(
                name=collection_name,
                namespace=collection_namespace,
                root=collection_root,
            ),
            test_path='tests',
            results_path='tests/output',
            sanity_path='tests/sanity',
            sanity_messages=sanity_messages,
            integration_path='tests/integration',
            integration_targets_path=integration_targets_path.rstrip(os.path.sep),
            integration_vars_path='tests/integration/integration_config.yml',
            integration_messages=integration_messages,
            unit_path='tests/unit',
            unit_module_path='tests/unit/plugins/modules',
            unit_module_utils_path='tests/unit/plugins/module_utils',
            unit_messages=unit_messages,
            unsupported=errors,
        )

    @staticmethod
    def __check_test_path(paths: list[str], messages: LayoutMessages) -> None:
        modern_test_path = 'tests/'
        modern_test_path_found = any(path.startswith(modern_test_path) for path in paths)
        legacy_test_path = 'test/'
        legacy_test_path_found = any(path.startswith(legacy_test_path) for path in paths)

        if modern_test_path_found and legacy_test_path_found:
            messages.warning.append('Ignoring tests in "%s" in favor of "%s".' % (legacy_test_path, modern_test_path))
        elif legacy_test_path_found:
            messages.warning.append('Ignoring tests in "%s" that should be in "%s".' % (legacy_test_path, modern_test_path))

    @staticmethod
    def __check_integration_path(paths: list[str], messages: LayoutMessages) -> str:
        modern_integration_path = 'roles/test/'
        modern_integration_path_found = any(path.startswith(modern_integration_path) for path in paths)
        legacy_integration_path = 'tests/integration/targets/'
        legacy_integration_path_found = any(path.startswith(legacy_integration_path) for path in paths)

        if modern_integration_path_found and legacy_integration_path_found:
            messages.warning.append('Ignoring tests in "%s" in favor of "%s".' % (legacy_integration_path, modern_integration_path))
            integration_targets_path = modern_integration_path
        elif legacy_integration_path_found:
            messages.info.append('Falling back to tests in "%s" because "%s" was not found.' % (legacy_integration_path, modern_integration_path))
            integration_targets_path = legacy_integration_path
        elif modern_integration_path_found:
            messages.info.append('Loading tests from "%s".' % modern_integration_path)
            integration_targets_path = modern_integration_path
        else:
            messages.error.append('Cannot run integration tests without "%s" or "%s".' % (modern_integration_path, legacy_integration_path))
            integration_targets_path = modern_integration_path

        return integration_targets_path

    @staticmethod
    def __check_unit_path(paths: list[str], messages: LayoutMessages) -> None:
        modern_unit_path = 'tests/unit/'
        modern_unit_path_found = any(path.startswith(modern_unit_path) for path in paths)
        legacy_unit_path = 'tests/units/'  # test/units/ will be covered by the warnings for test/ vs tests/
        legacy_unit_path_found = any(path.startswith(legacy_unit_path) for path in paths)

        if modern_unit_path_found and legacy_unit_path_found:
            messages.warning.append('Ignoring tests in "%s" in favor of "%s".' % (legacy_unit_path, modern_unit_path))
        elif legacy_unit_path_found:
            messages.warning.append('Rename "%s" to "%s" to run unit tests.' % (legacy_unit_path, modern_unit_path))
        elif modern_unit_path_found:
            pass  # unit tests only run from one directory so no message is needed
        else:
            messages.error.append('Cannot run unit tests without "%s".' % modern_unit_path)



# (c) 2017, Dag Wieers <dag@wieers.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# CI-required python3 boilerplate
from __future__ import annotations

import time
from datetime import datetime, timedelta, timezone

from ansible.module_utils.common.text.converters import to_text
from ansible.plugins.action import ActionBase
from ansible.utils.display import Display

display = Display()


class TimedOutException(Exception):
    pass


class ActionModule(ActionBase):
    TRANSFERS_FILES = False
    _VALID_ARGS = frozenset(('connect_timeout', 'delay', 'sleep', 'timeout'))

    DEFAULT_CONNECT_TIMEOUT = 5
    DEFAULT_DELAY = 0
    DEFAULT_SLEEP = 1
    DEFAULT_TIMEOUT = 600

    def do_until_success_or_timeout(self, what, timeout, connect_timeout, what_desc, sleep=1):
        max_end_time = datetime.now(timezone.utc) + timedelta(seconds=timeout)

        e = None
        while datetime.now(timezone.utc) < max_end_time:
            try:
                what(connect_timeout)
                if what_desc:
                    display.debug("wait_for_connection: %s success" % what_desc)
                return
            except Exception as e:
                error = e  # PY3 compatibility to store exception for use outside of this block
                if what_desc:
                    display.debug("wait_for_connection: %s fail (expected), retrying in %d seconds..." % (what_desc, sleep))
                time.sleep(sleep)

        raise TimedOutException("timed out waiting for %s: %s" % (what_desc, error))

    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()

        connect_timeout = int(self._task.args.get('connect_timeout', self.DEFAULT_CONNECT_TIMEOUT))
        delay = int(self._task.args.get('delay', self.DEFAULT_DELAY))
        sleep = int(self._task.args.get('sleep', self.DEFAULT_SLEEP))
        timeout = int(self._task.args.get('timeout', self.DEFAULT_TIMEOUT))

        if self._task.check_mode:
            display.vvv("wait_for_connection: skipping for check_mode")
            return dict(skipped=True)

        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        def ping_module_test(connect_timeout):
            """ Test ping module, if available """
            display.vvv("wait_for_connection: attempting ping module test")
            # re-run interpreter discovery if we ran it in the first iteration
            if self._discovered_interpreter_key:
                task_vars['ansible_facts'].pop(self._discovered_interpreter_key, None)
            # call connection reset between runs if it's there
            try:
                self._connection.reset()
            except AttributeError:
                pass

            ping_result = self._execute_module(module_name='ansible.legacy.ping', module_args=dict(), task_vars=task_vars)

            # Test module output
            if ping_result['ping'] != 'pong':
                raise Exception('ping test failed')

        start = datetime.now()

        if delay:
            time.sleep(delay)

        try:
            # If the connection has a transport_test method, use it first
            if hasattr(self._connection, 'transport_test'):
                self.do_until_success_or_timeout(self._connection.transport_test, timeout, connect_timeout, what_desc="connection port up", sleep=sleep)

            # Use the ping module test to determine end-to-end connectivity
            self.do_until_success_or_timeout(ping_module_test, timeout, connect_timeout, what_desc="ping module test", sleep=sleep)

        except TimedOutException as e:
            result['failed'] = True
            result['msg'] = to_text(e)

        elapsed = datetime.now() - start
        result['elapsed'] = elapsed.seconds

        # remove a temporary path we created
        self._remove_tmp_path(self._connection._shell.tmpdir)

        return result



from __future__ import annotations


class ModuleDocFragment:
    DOCUMENTATION = """
    options:
      fragment_expression:
        description: a fragment hosted expression that must be trusted whose default resolves to 4
        default: 2 + 2
    """



from __future__ import annotations


def do_nothing(myval):
    return myval


class FilterModule(object):
    """ Ansible core jinja2 filters """

    def filters(self):
        return {
            'filter2': do_nothing,
        }



"""Execute unit tests using pytest."""

from __future__ import annotations

import os
import sys
import typing as t

from ...constants import (
    CONTROLLER_MIN_PYTHON_VERSION,
    CONTROLLER_PYTHON_VERSIONS,
    REMOTE_ONLY_PYTHON_VERSIONS,
    SUPPORTED_PYTHON_VERSIONS,
)

from ...io import (
    write_text_file,
    make_dirs,
)

from ...util import (
    ANSIBLE_TEST_DATA_ROOT,
    display,
    is_subdir,
    str_to_version,
    SubprocessError,
    ANSIBLE_LIB_ROOT,
    ANSIBLE_TEST_TARGET_ROOT,
)

from ...util_common import (
    ResultType,
    handle_layout_messages,
    create_temp_dir,
)

from ...ansible_util import (
    ansible_environment,
    get_ansible_python_path,
)

from ...target import (
    walk_internal_targets,
    walk_units_targets,
)

from ...config import (
    UnitsConfig,
)

from ...coverage_util import (
    cover_python,
)

from ...data import (
    data_context,
)

from ...executor import (
    AllTargetsSkipped,
    Delegate,
    get_changes_filter,
)

from ...python_requirements import (
    install_requirements,
    post_install,
)

from ...content_config import (
    get_content_config,
)

from ...host_configs import (
    PosixConfig,
)

from ...provisioning import (
    prepare_profiles,
)

from ...pypi_proxy import (
    configure_pypi_proxy,
)

from ...host_profiles import (
    PosixProfile,
)


class TestContext:
    """Contexts that unit tests run in based on the type of content."""

    controller = 'controller'
    modules = 'modules'
    module_utils = 'module_utils'


def command_units(args: UnitsConfig) -> None:
    """Run unit tests."""
    handle_layout_messages(data_context().content.unit_messages)

    changes = get_changes_filter(args)
    require = args.require + changes
    include = walk_internal_targets(walk_units_targets(), args.include, args.exclude, require)

    paths = [target.path for target in include]

    content_config = get_content_config(args)
    supported_remote_python_versions = content_config.modules.python_versions

    if content_config.modules.controller_only:
        # controller-only collections run modules/module_utils unit tests as controller-only tests
        module_paths = []
        module_utils_paths = []
    else:
        # normal collections run modules/module_utils unit tests isolated from controller code due to differences in python version requirements
        module_paths = [path for path in paths if is_subdir(path, data_context().content.unit_module_path)]
        module_utils_paths = [path for path in paths if is_subdir(path, data_context().content.unit_module_utils_path)]

    controller_paths = sorted(path for path in set(paths) - set(module_paths) - set(module_utils_paths))

    remote_paths = module_paths or module_utils_paths

    test_context_paths = {
        TestContext.modules: module_paths,
        TestContext.module_utils: module_utils_paths,
        TestContext.controller: controller_paths,
    }

    if not paths:
        raise AllTargetsSkipped()

    targets = t.cast(list[PosixConfig], args.targets)
    target_versions: dict[str, PosixConfig] = {target.python.version: target for target in targets}
    skipped_versions = args.host_settings.skipped_python_versions
    warn_versions = []

    # requested python versions that are remote-only and not supported by this collection
    test_versions = [version for version in target_versions if version in REMOTE_ONLY_PYTHON_VERSIONS and version not in supported_remote_python_versions]

    if test_versions:
        for version in test_versions:
            display.warning(f'Skipping unit tests on Python {version} because it is not supported by this collection.'
                            f' Supported Python versions are: {", ".join(content_config.python_versions)}')

        warn_versions.extend(test_versions)

        if warn_versions == list(target_versions):
            raise AllTargetsSkipped()

    if not remote_paths:
        # all selected unit tests are controller tests

        # requested python versions that are remote-only
        test_versions = [version for version in target_versions if version in REMOTE_ONLY_PYTHON_VERSIONS and version not in warn_versions]

        if test_versions:
            for version in test_versions:
                display.warning(f'Skipping unit tests on Python {version} because it is only supported by module/module_utils unit tests.'
                                ' No module/module_utils unit tests were selected.')

            warn_versions.extend(test_versions)

            if warn_versions == list(target_versions):
                raise AllTargetsSkipped()

    if not controller_paths:
        # all selected unit tests are remote tests

        # requested python versions that are not supported by remote tests for this collection
        test_versions = [version for version in target_versions if version not in supported_remote_python_versions and version not in warn_versions]

        if test_versions:
            for version in test_versions:
                display.warning(f'Skipping unit tests on Python {version} because it is not supported by module/module_utils unit tests of this collection.'
                                f' Supported Python versions are: {", ".join(supported_remote_python_versions)}')

            warn_versions.extend(test_versions)

            if warn_versions == list(target_versions):
                raise AllTargetsSkipped()

    host_state = prepare_profiles(args, targets_use_pypi=True)  # units

    if args.delegate:
        raise Delegate(host_state=host_state, require=changes, exclude=args.exclude)

    test_sets = []

    if args.requirements_mode != 'skip':
        configure_pypi_proxy(args, host_state.controller_profile)  # units

    for version in SUPPORTED_PYTHON_VERSIONS:
        if version not in target_versions and version not in skipped_versions:
            continue

        test_candidates = []

        for test_context, paths in test_context_paths.items():
            if test_context == TestContext.controller:
                if version not in CONTROLLER_PYTHON_VERSIONS:
                    continue
            else:
                if version not in supported_remote_python_versions:
                    continue

            if not paths:
                continue

            env = ansible_environment(args)

            env.update(
                PYTHONPATH=get_units_ansible_python_path(args, test_context),
                ANSIBLE_CONTROLLER_MIN_PYTHON_VERSION=CONTROLLER_MIN_PYTHON_VERSION,
            )

            test_candidates.append((test_context, paths, env))

        if not test_candidates:
            continue

        if version in skipped_versions:
            display.warning("Skipping unit tests on Python %s because it could not be found." % version)
            continue

        target_profiles: dict[str, PosixProfile] = {profile.config.python.version: profile for profile in host_state.targets(PosixProfile)}
        target_profile = target_profiles[version]

        final_candidates = [(test_context, target_profile.python, paths, env) for test_context, paths, env in test_candidates]
        controller = any(test_context == TestContext.controller for test_context, python, paths, env in final_candidates)

        if args.requirements_mode != 'skip':
            install_requirements(args, target_profile, target_profile.python, ansible=controller, command=True, controller=False)  # units
        else:
            post_install(target_profile)

        test_sets.extend(final_candidates)

    if args.requirements_mode == 'only':
        sys.exit()

    for test_context, python, paths, env in test_sets:
        # When using pytest-mock, make sure that features introduced in Python 3.8 are available to older Python versions.
        # This is done by enabling the mock_use_standalone_module feature, which forces use of mock even when unittest.mock is available.
        # Later Python versions have not introduced additional unittest.mock features, so use of mock is not needed as of Python 3.8.
        # If future Python versions introduce new unittest.mock features, they will not be available to older Python versions.
        # Having the cutoff at Python 3.8 also eases packaging of ansible-core since no supported controller version requires the use of mock.
        #
        # NOTE: This only affects use of pytest-mock.
        #       Collection unit tests may directly import mock, which will be provided by ansible-test when it installs requirements using pip.
        #       Although mock is available for ansible-core unit tests, they should import unittest.mock instead.
        if str_to_version(python.version) < (3, 8):
            config_name = 'legacy.ini'
        else:
            config_name = 'default.ini'

        cmd = [
            'pytest',
            '-r', 'a',
            '-n', str(args.num_workers) if args.num_workers else 'auto',
            '--color', 'yes' if args.color else 'no',
            '-p', 'no:cacheprovider',
            '-c', os.path.join(ANSIBLE_TEST_DATA_ROOT, 'pytest', 'config', config_name),
            '--junit-xml', os.path.join(ResultType.JUNIT.path, 'python%s-%s-units.xml' % (python.version, test_context)),
            '--strict-markers',  # added in pytest 4.5.0
            '--rootdir', data_context().content.root,
            '--confcutdir', data_context().content.root,  # avoid permission errors when running from an installed version and using pytest >= 8
        ]  # fmt:skip

        if not data_context().content.collection:
            cmd.append('--durations=25')

        plugins = []

        if args.coverage:
            plugins.append('ansible_pytest_coverage')

        if data_context().content.collection:
            plugins.append('ansible_pytest_collections')

        plugins.append('ansible_forked')

        if plugins:
            env['PYTHONPATH'] += ':%s' % os.path.join(ANSIBLE_TEST_TARGET_ROOT, 'pytest/plugins')
            env['PYTEST_PLUGINS'] = ','.join(plugins)

        if args.collect_only:
            cmd.append('--collect-only')

        if args.verbosity:
            cmd.append('-' + ('v' * args.verbosity))

        cmd.extend(paths)

        display.info('Unit test %s with Python %s' % (test_context, python.version))

        try:
            cover_python(args, python, cmd, test_context, env, capture=False)
        except SubprocessError as ex:
            # pytest exits with status code 5 when all tests are skipped, which isn't an error for our use case
            if ex.status != 5:
                raise


def get_units_ansible_python_path(args: UnitsConfig, test_context: str) -> str:
    """
    Return a directory usable for PYTHONPATH, containing only the modules and module_utils portion of the ansible package.
    The temporary directory created will be cached for the lifetime of the process and cleaned up at exit.
    """
    if test_context == TestContext.controller:
        return get_ansible_python_path(args)

    try:
        cache = get_units_ansible_python_path.cache  # type: ignore[attr-defined]
    except AttributeError:
        cache = get_units_ansible_python_path.cache = {}  # type: ignore[attr-defined]

    python_path = cache.get(test_context)

    if python_path:
        return python_path

    python_path = create_temp_dir(prefix='ansible-test-')
    ansible_path = os.path.join(python_path, 'ansible')
    ansible_test_path = os.path.join(python_path, 'ansible_test')

    write_text_file(os.path.join(ansible_path, '__init__.py'), '', True)
    os.symlink(os.path.join(ANSIBLE_LIB_ROOT, 'module_utils'), os.path.join(ansible_path, 'module_utils'))

    if data_context().content.collection:
        # built-in runtime configuration for the collection loader
        make_dirs(os.path.join(ansible_path, 'config'))
        os.symlink(os.path.join(ANSIBLE_LIB_ROOT, 'config', 'ansible_builtin_runtime.yml'), os.path.join(ansible_path, 'config', 'ansible_builtin_runtime.yml'))

        # current collection loader required by all python versions supported by the controller
        write_text_file(os.path.join(ansible_path, 'utils', '__init__.py'), '', True)
        os.symlink(os.path.join(ANSIBLE_LIB_ROOT, 'utils', 'collection_loader'), os.path.join(ansible_path, 'utils', 'collection_loader'))

        # legacy collection loader required by all python versions not supported by the controller
        write_text_file(os.path.join(ansible_test_path, '__init__.py'), '', True)
        write_text_file(os.path.join(ansible_test_path, '_internal', '__init__.py'), '', True)
    elif test_context == TestContext.modules:
        # only non-collection ansible module tests should have access to ansible built-in modules
        os.symlink(os.path.join(ANSIBLE_LIB_ROOT, 'modules'), os.path.join(ansible_path, 'modules'))

    cache[test_context] = python_path

    return python_path



from __future__ import annotations

from ansible._internal._datatag._tags import VaultedValue
from ansible.module_utils._internal._datatag import AnsibleTagHelper


def test_vaulted_value_non_propagation_untagged_dst() -> None:
    vv = VaultedValue(ciphertext='')

    src_value = vv.tag('hello src')
    dst_value = 'hello dst'

    result = AnsibleTagHelper.tag_copy(src_value, dst_value)

    assert result is dst_value


def test_vaulted_value_non_propagation_tagged_dst() -> None:
    vv = VaultedValue(ciphertext='')
    vv_dst = VaultedValue(ciphertext='dst')

    src_value = vv.tag('hello src')
    dst_value = vv_dst.tag('hello dst')

    result = AnsibleTagHelper.tag_copy(src_value, dst_value)

    assert vv != vv_dst
    assert AnsibleTagHelper.tags(result) == {vv_dst}


def test_vaulted_value_propagation_untagged_dst() -> None:
    vv = VaultedValue(ciphertext='')

    src_value = vv.tag('hello')
    dst_value = 'hello'

    result = AnsibleTagHelper.tag_copy(src_value, dst_value)

    assert AnsibleTagHelper.tags(result) == {vv}


def test_vaulted_value_propagation_tagged_dst() -> None:
    vv = VaultedValue(ciphertext='')
    vv_dst = VaultedValue(ciphertext='dst')

    src_value = vv.tag('hello')
    dst_value = vv_dst.tag('hello')

    result = AnsibleTagHelper.tag_copy(src_value, dst_value)

    assert vv != vv_dst
    assert AnsibleTagHelper.tags(result) == {vv}



from __future__ import annotations

import pytest

from ansible import errors

from units.test_utils.controller.display import emits_warnings


@pytest.mark.parametrize("name", (
    "AnsibleFilterTypeError",
    "_AnsibleActionDone",
))
def test_deprecated(name: str) -> None:
    with emits_warnings(deprecation_pattern='is deprecated'):
        getattr(errors, name)


def test_deprecated_attribute_error() -> None:
    with pytest.raises(AttributeError):
        getattr(errors, 'bogus')



#!/usr/bin/python

from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule

module = AnsibleModule(argument_spec=dict())

module.exit_json(**{'tempdir': module._remote_tmp})



"""Layout provider for Ansible source."""

from __future__ import annotations

import os

from . import (
    ContentLayout,
    LayoutProvider,
)

from ...util import (
    ANSIBLE_SOURCE_ROOT,
    ANSIBLE_TEST_ROOT,
)


class AnsibleLayout(LayoutProvider):
    """Layout provider for Ansible source."""

    @staticmethod
    def is_content_root(path: str) -> bool:
        """Return True if the given path is a content root for this provider."""
        return os.path.isfile(os.path.join(path, 'pyproject.toml')) and os.path.isdir(os.path.join(path, 'test/lib/ansible_test'))

    def create(self, root: str, paths: list[str]) -> ContentLayout:
        """Create a Layout using the given root and paths."""
        plugin_paths = dict((p, os.path.join('lib/ansible/plugins', p)) for p in self.PLUGIN_TYPES)

        plugin_paths.update(
            modules='lib/ansible/modules',
            module_utils='lib/ansible/module_utils',
        )

        errors: list[str] = []

        if root != ANSIBLE_SOURCE_ROOT:
            errors.extend((
                f'Cannot test "{root}" with ansible-test from "{ANSIBLE_TEST_ROOT}".',
                '',
                f'Did you intend to run "{root}/bin/ansible-test" instead?',
            ))

        return ContentLayout(
            root,
            paths,
            plugin_paths=plugin_paths,
            collection=None,
            test_path='test',
            results_path='test/results',
            sanity_path='test/sanity',
            sanity_messages=None,
            integration_path='test/integration',
            integration_targets_path='test/integration/targets',
            integration_vars_path='test/integration/integration_config.yml',
            integration_messages=None,
            unit_path='test/units',
            unit_module_path='test/units/modules',
            unit_module_utils_path='test/units/module_utils',
            unit_messages=None,
            unsupported=errors,
        )



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# Copyright 2012, Seth Vidal <skvidal@fedoraproject.org>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from collections.abc import Mapping

from ansible.errors import AnsibleActionFail
from ansible.module_utils.six import string_types
from ansible.plugins.action import ActionBase
from ansible.parsing.utils.addresses import parse_address
from ansible.utils.display import Display
from ansible.utils.vars import combine_vars

display = Display()


class ActionModule(ActionBase):
    """ Create inventory hosts and groups in the memory inventory"""

    # We need to be able to modify the inventory
    BYPASS_HOST_LOOP = True
    _requires_connection = False
    _supports_check_mode = True

    def run(self, tmp=None, task_vars=None):

        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        args = self._task.args
        raw = args.pop('_raw_params', {})
        if isinstance(raw, Mapping):
            # TODO: create 'conflict' detection in base class to deal with repeats and aliases and warn user
            args = combine_vars(raw, args)
        else:
            raise AnsibleActionFail('Invalid raw parameters passed, requires a dictionary/mapping got a  %s' % type(raw))

        # Parse out any hostname:port patterns
        new_name = args.get('name', args.get('hostname', args.get('host', None)))
        if new_name is None:
            raise AnsibleActionFail('name, host or hostname needs to be provided')

        display.vv("creating host via 'add_host': hostname=%s" % new_name)

        try:
            name, port = parse_address(new_name, allow_ranges=False)
        except Exception:
            # not a parsable hostname, but might still be usable
            name = new_name
            port = None

        if port:
            args['ansible_ssh_port'] = port

        groups = args.get('groupname', args.get('groups', args.get('group', '')))
        # add it to the group if that was specified
        new_groups = []
        if groups:
            if isinstance(groups, list):
                group_list = groups
            elif isinstance(groups, string_types):
                group_list = groups.split(",")
            else:
                raise AnsibleActionFail("Groups must be specified as a list.", obj=groups)

            for group_name in group_list:
                if group_name not in new_groups:
                    new_groups.append(group_name.strip())

        # Add any variables to the new_host
        host_vars = dict()
        special_args = frozenset(('name', 'hostname', 'groupname', 'groups'))
        for k in args.keys():
            if k not in special_args:
                host_vars[k] = args[k]

        result['changed'] = False
        result['add_host'] = dict(host_name=name, groups=new_groups, host_vars=host_vars)
        return result



from __future__ import annotations

import re
import pathlib
import sys
import typing as t

from ansible.module_utils._internal import _stack, _messages, _validation, _plugin_info


def deprecator_from_collection_name(collection_name: str | None) -> _messages.PluginInfo | None:
    """Returns an instance with the special `collection` type to refer to a non-plugin or ambiguous caller within a collection."""
    # CAUTION: This function is exposed in public API as ansible.module_utils.datatag.deprecator_from_collection_name.

    if not collection_name:
        return None

    _validation.validate_collection_name(collection_name)

    return _messages.PluginInfo(
        resolved_name=collection_name,
        type=None,
    )


def get_best_deprecator(*, deprecator: _messages.PluginInfo | None = None, collection_name: str | None = None) -> _messages.PluginInfo:
    """Return the best-available `PluginInfo` for the caller of this method."""
    _skip_stackwalk = True

    if deprecator and collection_name:
        raise ValueError('Specify only one of `deprecator` or `collection_name`.')

    return deprecator or deprecator_from_collection_name(collection_name) or get_caller_plugin_info() or INDETERMINATE_DEPRECATOR


def get_caller_plugin_info() -> _messages.PluginInfo | None:
    """Try to get `PluginInfo` for the caller of this method, ignoring marked infrastructure stack frames."""
    _skip_stackwalk = True

    if frame_info := _stack.caller_frame():
        return _path_as_plugininfo(frame_info.filename)

    return None  # pragma: nocover


def _path_as_plugininfo(path: str) -> _messages.PluginInfo | None:
    """Return a `PluginInfo` instance if the provided `path` refers to a plugin."""
    return _path_as_core_plugininfo(path) or _path_as_collection_plugininfo(path)


def _path_as_core_plugininfo(path: str) -> _messages.PluginInfo | None:
    """Return a `PluginInfo` instance if the provided `path` refers to a core plugin."""
    try:
        relpath = str(pathlib.Path(path).relative_to(_ANSIBLE_MODULE_BASE_PATH))
    except ValueError:
        return None  # not ansible-core

    namespace = 'ansible.builtin'

    if match := re.match(r'plugins/(?P<plugin_type>\w+)/(?P<plugin_name>\w+)', relpath):
        plugin_name = match.group("plugin_name")
        plugin_type = _plugin_info.normalize_plugin_type(match.group("plugin_type"))

        if plugin_type not in _DEPRECATOR_PLUGIN_TYPES:
            # The plugin type isn't a known deprecator type, so we have to assume the caller is intermediate code.
            # We have no way of knowing if the intermediate code is deprecating its own feature, or acting on behalf of another plugin.
            # Callers in this case need to identify the deprecating plugin name, otherwise only ansible-core will be reported.
            # Reporting ansible-core is never wrong, it just may be missing an additional detail (plugin name) in the "on behalf of" case.
            return ANSIBLE_CORE_DEPRECATOR

        if plugin_name == '__init__':
            # The plugin type is known, but the caller isn't a specific plugin -- instead, it's core plugin infrastructure (the base class).
            return _messages.PluginInfo(resolved_name=namespace, type=plugin_type)
    elif match := re.match(r'modules/(?P<module_name>\w+)', relpath):
        # AnsiballZ Python package for core modules
        plugin_name = match.group("module_name")
        plugin_type = _messages.PluginType.MODULE
    elif match := re.match(r'legacy/(?P<module_name>\w+)', relpath):
        # AnsiballZ Python package for non-core library/role modules
        namespace = 'ansible.legacy'

        plugin_name = match.group("module_name")
        plugin_type = _messages.PluginType.MODULE
    else:
        return ANSIBLE_CORE_DEPRECATOR  # non-plugin core path, safe to use ansible-core for the same reason as the non-deprecator plugin type case above

    name = f'{namespace}.{plugin_name}'

    return _messages.PluginInfo(resolved_name=name, type=plugin_type)


def _path_as_collection_plugininfo(path: str) -> _messages.PluginInfo | None:
    """Return a `PluginInfo` instance if the provided `path` refers to a collection plugin."""
    if not (match := re.search(r'/ansible_collections/(?P<ns>\w+)/(?P<coll>\w+)/plugins/(?P<plugin_type>\w+)/(?P<plugin_name>\w+)', path)):
        return None

    plugin_type = _plugin_info.normalize_plugin_type(match.group('plugin_type'))

    if plugin_type in _AMBIGUOUS_DEPRECATOR_PLUGIN_TYPES:
        # We're able to detect the namespace, collection and plugin type -- but we have no way to identify the plugin name currently.
        # To keep things simple we'll fall back to just identifying the namespace and collection.
        # In the future we could improve the detection and/or make it easier for a caller to identify the plugin name.
        return deprecator_from_collection_name('.'.join((match.group('ns'), match.group('coll'))))

    if plugin_type not in _DEPRECATOR_PLUGIN_TYPES:
        # The plugin type isn't a known deprecator type, so we have to assume the caller is intermediate code.
        # We have no way of knowing if the intermediate code is deprecating its own feature, or acting on behalf of another plugin.
        # Callers in this case need to identify the deprecator to avoid ambiguity, since it could be the same collection or another collection.
        return INDETERMINATE_DEPRECATOR

    name = '.'.join((match.group('ns'), match.group('coll'), match.group('plugin_name')))

    # DTFIX-FUTURE: deprecations from __init__ will be incorrectly attributed to a plugin of that name

    return _messages.PluginInfo(resolved_name=name, type=plugin_type)


_ANSIBLE_MODULE_BASE_PATH: t.Final = pathlib.Path(sys.modules['ansible'].__file__).parent
"""Runtime-detected base path of the `ansible` Python package to distinguish between Ansible-owned and external code."""

ANSIBLE_CORE_DEPRECATOR: t.Final = deprecator_from_collection_name('ansible.builtin')
"""Singleton `PluginInfo` instance for ansible-core callers where the plugin can/should not be identified in messages."""

INDETERMINATE_DEPRECATOR: t.Final = _messages.PluginInfo(resolved_name=None, type=None)
"""Singleton `PluginInfo` instance for indeterminate deprecator."""

_DEPRECATOR_PLUGIN_TYPES: t.Final = frozenset(
    {
        _messages.PluginType.ACTION,
        _messages.PluginType.BECOME,
        _messages.PluginType.CACHE,
        _messages.PluginType.CALLBACK,
        _messages.PluginType.CLICONF,
        _messages.PluginType.CONNECTION,
        # DOC_FRAGMENTS - no code execution
        # FILTER - basename inadequate to identify plugin
        _messages.PluginType.HTTPAPI,
        _messages.PluginType.INVENTORY,
        _messages.PluginType.LOOKUP,
        _messages.PluginType.MODULE,  # only for collections
        _messages.PluginType.NETCONF,
        _messages.PluginType.SHELL,
        _messages.PluginType.STRATEGY,
        _messages.PluginType.TERMINAL,
        # TEST - basename inadequate to identify plugin
        _messages.PluginType.VARS,
    }
)
"""Plugin types which are valid for identifying a deprecator for deprecation purposes."""

_AMBIGUOUS_DEPRECATOR_PLUGIN_TYPES: t.Final = frozenset(
    {
        _messages.PluginType.FILTER,
        _messages.PluginType.TEST,
    }
)
"""Plugin types for which basename cannot be used to identify the plugin name."""
